from typing import Protocol, runtime_checkable
from redis.asyncio.client import Pipeline


@runtime_checkable
class ExecuteAbleQuery(Protocol):
    async def execute(self):
        raise NotImplementedError


class LazySession(Pipeline):
    async def execute(self):
        return await super().execute()

    async def commit(self):
        return await super().execute()

    async def rollback(self):
        return await super().discard()

    async def close(self):
        return await self.__aexit__(None, None, None)




from typing import Generic
from .types import TIModel, TOModel
from .adder import AdderRepo
from .getter import GetterRepo
from .updateter import UpdateterRepo
from .deleter import DeleterRepo
from .paginated import PaginatedRepo
from .status_updater import StatusUpdaterRepo
from .paginated_all_getter import PaginatedAllGetterRepo
from .base_setter import BaseSetterRepo


class BaseRepository(
    BaseSetterRepo[TIModel, TOModel], 
    AdderRepo[TIModel, TOModel], 
    GetterRepo[TIModel, TOModel], 
    UpdateterRepo[TIModel, TOModel], 
    DeleterRepo[TIModel, TOModel],
    PaginatedRepo[TIModel, TOModel],
    StatusUpdaterRepo[TIModel, TOModel],
    PaginatedAllGetterRepo[TIModel, TOModel],
    Generic[TIModel, TOModel],
):
    __abstract__ = True




from typing import Generic, List, Optional
from .types import TIModel, TOModel
from .all_getter import AllGetterRepo


class PaginatedAllGetterRepo(AllGetterRepo[TIModel, TOModel], Generic[TIModel, TOModel]):
    __abstract__ = True

    def _page(self, count, page, objs):
        if count and page:
            start_point = (count or 1) * page - count
            end_point = start_point + count
            return objs[start_point:end_point]

        return objs

    async def page(
        self, 
        filters: Optional[str] = None, 
        count: Optional[int] = None, 
        page: Optional[int] = None,
        *,
        parse: bool = True,
    ) -> List[TOModel]:
        objs = await super().all(filters, parse=parse)
        return self._page(count, page, objs)




import json
from typing import Generic
from .types import TIModel, TOModel
from .base import BaseRepoCreator
from redis.commands.json.path import Path


class UpdateterRepo(BaseRepoCreator[TIModel, TOModel], Generic[TIModel, TOModel]):
    __abstract__ = True

    async def update(self, obj: TIModel):
        id = self.get_obj_id(obj)
        return self.session.json().set(
            self.hname, 
            Path(f'$.{id}').strPath,
            json.loads(obj.json()),
        )




from typing import Generic, Union
from .types import TIModel, TOModel
from .base import BaseRepoCreator
from redis.commands.json.path import Path
from redis.exceptions import ResponseError


class GetterRepo(BaseRepoCreator[TIModel, TOModel], Generic[TIModel, TOModel]):
    __abstract__ = True

    async def get(self, id, filters: str = '') -> Union[TOModel, None]: 
        try: 
            objs = await self.session.json().get(
                self.hname, 
                Path(f'.{id}{filters}').strPath,
            ).execute()  # type: ignore
        except ResponseError as e:
            return print(str(e))

        obj = self._unwrap_till(objs, condition_func=self._is_non_empty_list)
        return obj and self.OSchema(**obj)



from typing import Generic
from redis.commands.json.path import Path
from .types import TIModel, TOModel
from .base import BaseRepoCreator


class StatusUpdaterRepo(BaseRepoCreator[TIModel, TOModel], Generic[TIModel, TOModel]):
    __abstract__ = True

    async def deactivate(self, id):
        return self.update_status(id, status=False)

    async def update_status(self, id, status: bool): 
        return self.session.json().set(
            self.hname, 
            Path(f'$.{id}.is_active').strPath,
            status,
        )




from typing import Generic
from .types import TIModel, TOModel
from .base import BaseRepoCreator
from redis.commands.json.path import Path


class BaseSetterRepo(BaseRepoCreator[TIModel, TOModel], Generic[TIModel, TOModel]):
    __abstract__ = True

    async def set_base(self, obj: dict):
        self.session.json().set(self.hname, Path('$').strPath, obj)
        return obj

    async def extend_base(self, obj: dict[str, dict]):
        for k, v in obj.items():
            self.session.json().set(self.hname, Path(str(k)).strPath, v)

        return obj




import json
from typing import Any, Callable, Optional, Type, Generic, Union
from uuid import uuid4
from axabc.db.async_repository import AbstractAsyncRepository
from async_redis_uow.session_maker.session import LazySession 
from .types import TIModel, TOModel


class BaseRepoCreator(AbstractAsyncRepository, Generic[TIModel, TOModel]):
    Schema: Type[TIModel]
    OSchema: Type[TOModel]

    __hname__: Optional[str] = None
    __abstract__ = True

    def __init__(self, session: LazySession) -> None:
        self.session = session

    def __init_subclass__(cls) -> None:
        if cls.__abstract__ and '__abstract__' in cls.__dict__:
            return

        types = getattr(cls, "__orig_bases__")[0].__args__
        cls.Schema, cls.OSchema = types

    def get_obj_id(self, obj: Union[TIModel, TOModel]):
        return _ if hasattr(obj, 'id') and (_ := getattr(obj, 'id')) else str(uuid4())

    @staticmethod
    def _isnt_last_level_list(objs: list[TIModel]):
        return objs and len(objs) == 1 and isinstance(objs[-1], list)

    @staticmethod
    def _is_non_empty_list(objs: list[TIModel]):
        return isinstance(objs, list) and objs
    
    def _unwrap_till(self, objs, *, condition_func: Callable[[list[TIModel]], Any] = _isnt_last_level_list):
        while condition_func(objs):
            objs = objs[-1]

        return objs

    @property
    def hname(self):
        return self.__hname__ or self.Schema.__name__.lower()

    def dumps(self, obj: dict) -> str:
        return json.dumps(obj)

    def loads(self, obj: str) -> dict:
        return json.loads(obj)




from typing import Generic
from .types import TIModel, TOModel
from .base import BaseRepoCreator
from redis.commands.json.path import Path


class DeleterRepo(BaseRepoCreator[TIModel, TOModel], Generic[TIModel, TOModel]):
    __abstract__ = True

    async def delete(self, id):
        return self.session.json().delete(
            self.hname,
            Path(f'$.{id}').strPath,
        )




from redis.commands.json.path import Path
from axabc.db import AsyncUOWFactory
from async_redis_uow.repository.common import BaseRepository
from async_redis_uow.session_maker.session import LazySession
from async_redis_uow.session_maker.session import LazySession


async def create_models(uowf: AsyncUOWFactory):
    async with uowf() as uow:
        for name in uow.repo.get_repos():
            repo = getattr(uow.repo, name)
            if not isinstance(repo, BaseRepository):
                continue

            repo_session: LazySession = repo.session

            if not isinstance(repo_session, LazySession):  
                continue

            res = await repo_session.json().get(repo.hname).execute()  # type: ignore
            if res is None or res == [None]:
                await repo_session.json().set(repo.hname, Path('$'), {}).execute()  # type: ignore




from typing import Optional
from redis.asyncio.client import Redis
from axabc.db.session_mapper import LazySessionMaker as _LazySessionMaker

from .session import LazySession


class RedisLazySessionMaker(_LazySessionMaker, Redis):
    def __call__(self, transaction: bool = True, shard_hint: Optional[str] = None) -> LazySession:
        return LazySession(self.connection_pool, self.response_callbacks, transaction, shard_hint)  # type: ignore




from typing import Any, Callable, Generic, List, Optional
from pydantic import parse_obj_as
from redis.commands.json.path import Path
from redis.exceptions import ResponseError
from .types import TIModel, TOModel
from .base import BaseRepoCreator


class AllGetterRepo(BaseRepoCreator[TIModel, TOModel], Generic[TIModel, TOModel]):
    __abstract__ = True

    def _all_sort_key(self, obj: TOModel):
        return obj.created_at

    async def all(
        self, 
        filters: Optional[str] = None, 
        *, 
        parse: bool = True, 
        sort_func: Optional[Callable[[TOModel], Any]] = None,
    ) -> List[TOModel]:
        filters = filters or '$.[*]'

        try:
            objs = await self.session.json().get(
                self.hname, 
                Path(f'{filters}').strPath,
            ).execute()  # type: ignore
        except ResponseError:
            return []
        
        objs = self._unwrap_till(objs)

        if parse:
            sort_func = sort_func or self._all_sort_key
            return sorted(parse_obj_as(List[self.OSchema], objs), key=sort_func)
        
        return objs




from typing import Generic, Iterable, Optional
from axsqlalchemy.repository.paginated import math
from redis.commands.json.path import Path
from .types import TIModel, TOModel
from .all_getter import AllGetterRepo


class PaginatedRepo(AllGetterRepo[TIModel, TOModel], Generic[TIModel, TOModel]):
    __abstract__ = True

    async def all_page_count(self, filters: str = '', count: Optional[int] = None) -> int:
        all_count = await self.all_count(filters)
        return math.ceil(all_count / count) if count else 1 

    async def all_count(self, filters: str = ''):
        res = await self.session.json().objlen(
            self.hname, 
            Path(f'${filters}').strPath,
        ).execute()  # type: ignore

        return res[-1][-1] if res and res[-1] and isinstance(res[-1], Iterable) else 0




from .adder import AdderRepo
from .getter import GetterRepo
from .updateter import UpdateterRepo
from .deleter import DeleterRepo
from .paginated import PaginatedRepo
from .status_updater import StatusUpdaterRepo
from .paginated_all_getter import PaginatedAllGetterRepo
from .common import BaseRepository


__all__ = [
	'AdderRepo',
	'GetterRepo',
	'UpdateterRepo',
	'DeleterRepo',
	'PaginatedRepo',
	'StatusUpdaterRepo',
	'PaginatedAllGetterRepo',
	'BaseRepository',
]

